//Extension of simple_ls example to allow copying of files from PC to uSD
//By Raymond Allen, MIT license 2020

#include "propeller2.h"

#define USE_HOST  //Using Plan9

#include "ff.h"
#include "diskio.h"

#include <stdio.h>
#include <unistd.h>
#include <dirent.h>
#include <sys/stat.h>


BYTE Buff[4096];	/* Working buffer */
FATFS FatFs[2];		/* Filesystem object for each logical drive */
FIL File[2];		/* File object */
FFDIR Dir;			/* Directory object */  //RJA:  Fastspin renames DIR to FFDIR
FILINFO Finfo;
DWORD AccSize;				/* Work register for fs command */
WORD AccFiles, AccDirs;


/*--------------------------------------------------------------------*/
/* User Provided Timer Function for FatFs module                      */
/*--------------------------------------------------------------------*/
/* This is a real time clock service to be called from FatFs module.  */
/* Any valid time must be returned even if the system does not        */
/* support a real time clock. This is not required when FatFs is      */
/* configured for FF_FS_READONLY or FF_FS_NORTC = 1.                  */
/*--------------------------------------------------------------------*/


DWORD get_fattime(void)  //Note:  RJA set FF_FS_NORTC to 0 in ffconf.h to enable RTC support
{
    return 0;
    //RTC rtc;


    //if (!RtcOk) return 0;

    ///* Get local time */
    //rtc_gettime(&rtc);

    ///* Pack date and time into a DWORD variable */
    //return	  ((DWORD)(rtc.year - 1980) << 25)
    //    | ((DWORD)rtc.month << 21)
    //    | ((DWORD)rtc.mday << 16)
    //    | ((DWORD)rtc.hour << 11)
    //    | ((DWORD)rtc.min << 5)
    //    | ((DWORD)rtc.sec >> 1);
}

static
void put_rc(FRESULT rc)
{//RJA had to rewrite this due to fastspin compiler error
    const char* p;
    switch (rc)
    {
    case 0:
        p = "OK"; break;
    case 1:
        p = "DISK_ERR"; break;
        break;
    case 2:
        p = "INT_ERR"; break;
    case 3:
        p = "NOT_READY"; break;
    case 4:
        p = "NO_FILE"; break;
    case 5:
        p = "NO_PATH"; break;
    case 6:
        p = "INVALID_NAME"; break;
    case 7:
        p = "DENIED"; break;
    case 8:
        p = "EXIST"; break;
    case 9:
        p = "INVALID_OBJECT"; break;
    case 10:
        p = "WRITE_PROTECTED"; break;
    case 11:
        p = "INVALID_DRIVE"; break;
    case 12:
        p = "NOT_ENABLED"; break;
    case 13:
        p = "NO_FILE_SYSTEM"; break;
    case 14:
        p = "MKFS_ABORTED"; break;
    case 15:
        p = "TIMEOUT"; break;
    case 16:
        p = "LOCKED"; break;
    case 17:
        p = "NOT_ENOUGH_CORE"; break;
    case 18:
        p = "TOO_MANY_OPEN_FILES"; break;
    default:
        p = "Unknown Error.";

    }
    printf("rc=%u FR_%S\n", rc, p);
}


void ListDir(char* ptr)
{//Show contents of given folder on uSD card
    FFDIR Dir1;
    FRESULT fr;
    int p1, s1, s2;
    FATFS* fs;

    fr = f_opendir(&Dir1, ptr);
    printf("uSD Dir %s open with result %d\n", ptr, fr);
    put_rc(fr);
    if (fr) {
        return;
    }
    p1 = s1 = s2 = 0;
    for (;;) {
        fr = f_readdir(&Dir1, &Finfo);
        if ((fr != FR_OK) || !Finfo.fname[0])
        {
            //RJA you get here after the last directory entry...
            return;
        }
        if (Finfo.fattrib & AM_DIR) {
            s2++;
        }
        else {
            s1++; p1 += Finfo.fsize;
        }
        printf("%c%c%c%c%c %u/%02u/%02u %02u:%02u %9lu  %s\n",
            (Finfo.fattrib & AM_DIR) ? 'D' : '-',
            (Finfo.fattrib & AM_RDO) ? 'R' : '-',
            (Finfo.fattrib & AM_HID) ? 'H' : '-',
            (Finfo.fattrib & AM_SYS) ? 'S' : '-',
            (Finfo.fattrib & AM_ARC) ? 'A' : '-',
            (Finfo.fdate >> 9) + 1980, (Finfo.fdate >> 5) & 15, Finfo.fdate & 31,
            (Finfo.ftime >> 11), (Finfo.ftime >> 5) & 63,
            (DWORD)Finfo.fsize,
            Finfo.fname);
    }
}


int main()
{//Show Plan9 and uSD folder contents and then copy a file from computer to uSD

    //First, mount Plan9 and then show root folder contents
    //Vars for Plan9
    DIR *d;
    struct dirent *ent;
    struct stat sbuf;
    int r;

    //Mount Plan9 computer folder
    mount("/files", _vfs_open_host());
    chdir("/files");
    printf("Plan9 mounted.\n");

    d = opendir(".");
    if (!d) {
        printf("Unable to open Plan9 root directory on computer.\n");
        return 1;
    }
    printf("Plan9root directory opened.\n");
    
    printf("Files in Plan9 current directory:\n");
    for(;;) {
        ent = readdir(d);
        if (!ent) break;
        r = stat(ent->d_name, &sbuf);
        if (r) {
            printf("??? error in stat of %s\n", ent->d_name);
        } else if ((sbuf.st_mode & S_IFMT) == S_IFDIR) {
            printf("%8s %s\n", "<dir>", ent->d_name);
        } else {
            printf("%8u %s\n", sbuf.st_size, ent->d_name);
        }
    }
    closedir(d);


    //Next, mount uSD and show it's root folder contents
    //Vars for uSD access
    UINT bw;
    FRESULT fr;
    FIL Fil;			/* File object needed for each open file */
    printf("\n\n\nMounting uSD.\n");
    sleep(1);
    for (int i = 1; i < 10; i++)
    {
        printf("trying to mount...\n");
        fr = f_mount(&FatFs, "/", 1);		/* Give a work area to the default drive */
        printf("Result of mount= %d\n", fr);
        put_rc(fr);
        if (fr == 0)
            break;
        sleep(1);
    }
    if (fr != 0)
    {
        printf("Failed to mount uSD...");
        return 1;
    }
    printf("\n\nuSD root directory contents:\n");
    ListDir(".");

    //Finally copy a file from computer to uSD
    char* filename = "text.txt";
    printf("\n\nCopying file %s from computer to uSD.\n", filename);
    FILE* f;
    f = fopen("test.txt", "r");
    if (!f) {
        printf("Unable to open test.txt on computer's Plan9 root folder.\n");
        return 1;
    }
    printf("File 'Text.txt' on computer opened.\n");
    char filemode = FA_WRITE | FA_CREATE_ALWAYS;
    printf("Opening file %s on uSD card for writing.\n", filename);
    fr=f_open(&File[0], filename, filemode);
    printf("Result of opening file for write= %d\n", fr);
    put_rc(fr);
    if (fr != 0)
    {
        printf("Unable to open uSD file for writing.\n");
        return 1;
    }
    //Now, the copy
    printf("\n\nCopying file from computer to uSD...\n");
    char inbuffer[1000];
    int nread = 1000;
    int nwritten;
    while (nread ==1000)
    {//read and write until done
        nread=fread(inbuffer, 1, 1000, f);
        fr=f_write(&File[0], inbuffer, nread, &nwritten);
        if ((nread != nwritten) || (fr != 0))
        {
            printf("Error writing to uSD.");
            return 1;
        }
    }
    printf("Copy from Plan9 to uSD complete.\n");
    //close both files
    fclose(f); //close Plan9 file
    f_close(&File[0]);  //close uSD file
    printf("Input and output files closed\n");

    return 0;

}
